import pytest
from unittest.mock import call


@pytest.fixture
def shared_kpbig(shared_kp_factory):
    data = shared_kp_factory("db_argon2_kdbx4_pass_big.kdbx", password="foobar")
    data.ioh.print.reset_mock()
    return data


full_ls = [
    call("foo"),
    call("group/entry"),
    call("group/subgroup/entry"),
    call("group/subgroup/other entry"),
    call("other/foo"),
]

group_ls = [
    call("group/entry"),
    call("group/subgroup/entry"),
    call("group/subgroup/other entry"),
]


@pytest.mark.parametrize(
    "glob,expected_prints",
    [
        ("", full_ls),
        ("*", full_ls),
        ("foo", [call("foo")]),
        ("foobar", []),
        ("f*", [call("foo")]),
        ("group/*", group_ls),
    ],
)
def test_ls_glob(shared_kpbig, glob, expected_prints, th):
    th.cmd(f"ls {glob}", shared_kpbig.kp, shared_kpbig.ioh)
    shared_kpbig.ioh.print.assert_has_calls(expected_prints)


def test_show(shared_kpbig, th):
    th.cmd("show foo", shared_kpbig.kp, shared_kpbig.ioh)
    shared_kpbig.ioh.print.assert_has_calls(
        [
            call("path: foo"),
            call("username: myuser"),
            call("password: mypass"),
            call("url: http://example.com"),
            call("autotype_sequence: {USER}{PASSWORD}"),
            call("notes[1]: first note"),
            call("notes[2]: second note"),
        ]
    )


def test_show_n(shared_kpbig, th):
    th.cmd("show foo -n", shared_kpbig.kp, shared_kpbig.ioh)
    shared_kpbig.ioh.print.assert_has_calls(
        [
            call("foo"),
            call("myuser"),
            call("mypass"),
            call("http://example.com"),
            call("{USER}{PASSWORD}"),
            call("first note"),
            call("second note"),
        ]
    )


@pytest.mark.parametrize(
    "field,lines",
    [
        ("path", ["foo"]),
        ("username", ["myuser"]),
        ("password", ["mypass"]),
        ("url", ["http://example.com"]),
        ("autotype_sequence", ["{USER}{PASSWORD}"]),
        ("notes", ["first note", "second note"]),
    ],
)
def test_show_field(shared_kpbig, field, lines, th):
    th.cmd(f"show foo {field}", shared_kpbig.kp, shared_kpbig.ioh)
    expected_calls = (
        [call(f"{field}: {lines[0]}")]
        if len(lines) == 1
        else [call(f"{field}[{i+1}]: {line}") for i, line in enumerate(lines)]
    )
    shared_kpbig.ioh.print.assert_has_calls(expected_calls)

    shared_kpbig.ioh.print.reset_mock()
    th.cmd(f"show -n foo {field}", shared_kpbig.kp, shared_kpbig.ioh)
    expected_calls = [call(line) for line in lines]
    shared_kpbig.ioh.print.assert_has_calls(expected_calls)


def test_show_many_fields_different_order(shared_kpbig, th):
    th.cmd(f"show foo url path", shared_kpbig.kp, shared_kpbig.ioh)
    shared_kpbig.ioh.print.assert_has_calls(
        [call("url: http://example.com"), call("path: foo")]
    )


def test_db(shared_kpbig, th):
    th.cmd("db", shared_kpbig.kp, shared_kpbig.ioh)
    shared_kpbig.ioh.print.assert_has_calls(
        [call(shared_kpbig.args.db), call("Locked: False")]
    )


def test_echo(shared_kpbig, th):
    msg = "foo bar baz"
    th.cmd(f"echo '{msg}'", shared_kpbig.kp, shared_kpbig.ioh)
    shared_kpbig.ioh.print.assert_has_calls([call(msg)])



import os
import shutil
import re
import shlex
from functools import cache, partial
from collections import namedtuple

import pytest
from unittest.mock import Mock

from kpsh.commands import prepare_command_parser, CommandError
from kpsh.db import DelayedPyKeePass
from kpsh.io import Handler


class ProgrammableIO(Handler):
    def __init__(self):
        self.prompt = Mock()
        self.print = Mock()
        self.eprint = Mock()
        self.get_command = Mock()
        super().__init__()


class Args:
    def __getattr__(self, name):
        return self.__dict__.get(name, None)


def copy_db_data(dbname, destdir):
    curdir = os.path.dirname(os.path.abspath(__file__))

    keysrc = os.path.join(curdir, "db", "keyfile")
    keydst = os.path.join(destdir, "keyfile")
    shutil.copy(keysrc, keydst)

    src = os.path.join(curdir, "db", dbname)
    dst = os.path.join(destdir, dbname)
    shutil.copy(src, dst)
    return dst


def make_kpdata(**kw):
    Data = namedtuple("Data", ("kp", "args", "ioh"))

    iohandler = ProgrammableIO()

    args = Args()
    args.__dict__.update(kw)

    kp = DelayedPyKeePass(args, iohandler)

    return Data(kp, args, iohandler)


@pytest.fixture
def iohandler():
    return ProgrammableIO()


@pytest.fixture
def fakeargs():
    return Args()


@pytest.fixture
def tmpdb(tmpdir):
    factory = partial(copy_db_data, destdir=str(tmpdir))
    return factory


@pytest.fixture
def kp_factory(tmpdb):
    def factory(dbpath, **args):
        db = tmpdb(dbpath)
        return make_kpdata(db=db, **args)

    return factory


@pytest.fixture
def kp_open():
    def opener(path, **args):
        return make_kpdata(db=path, **args)

    return opener


@pytest.fixture(scope="session")
def shared_tmpdb(tmpdir_factory):
    tmpdir = tmpdir_factory.mktemp("databases")
    factory = partial(copy_db_data, destdir=tmpdir)
    return factory


# opening databases is slow... so we introduce a shared DelayedPyKeePass, which
# can be accessed by a few different tests. Generally it should be used only to
# test read-only operations (not counting unlocking, because usually it doesn't
# hurt that database has been unlocked in some other test)
@pytest.fixture(scope="session")
def shared_kp_factory(shared_tmpdb):
    @cache
    def factory(dbpath, **args):
        db = shared_tmpdb(dbpath)
        return make_kpdata(db=db, **args)

    return factory


class Helpers:
    @staticmethod
    def cmd(text, kp, ioh):
        parser, _ = prepare_command_parser()
        text_spl = shlex.split(text) if isinstance(text, str) else text
        args = parser.parse_args(text_spl)
        return args.func(kp, args, ioh)

    @staticmethod
    def fields(path, kp):
        fields = {}

        def store_field(arg):
            key, _, val = arg.partition(":")

            # examples of keys: path, notes[0]
            m = re.match(r"([^[]+)(?:\[(\d+)\])?", key)

            key = m[1].strip()
            id = m[2]
            val = val[1:]  # strip the initial space added by kpsh

            if id is not None:
                vals = fields.setdefault(key, [])
                vals.append(val)
                assert int(id) == len(vals)
            else:
                fields[key] = val

        ioh = ProgrammableIO()
        ioh.print.side_effect = store_field

        try:
            Helpers.cmd(f"show {path}", kp, ioh)
        except CommandError:
            return None

        return fields

    @staticmethod
    def assert_fields(kp, path, **fields):
        fields.setdefault("path", path)
        existing_fields = Helpers.fields(path, kp)
        assert fields is not None, f"entry '{path}' doesn't exist in database"
        assert fields == existing_fields


@pytest.fixture
def th():
    return Helpers



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2019 Michał Góral.

import time
import re
import subprocess


def _token_delay(token):
    match = re.match(r"{DELAY (\d+)}", token)
    if match:
        delay = match.group(1)
        return lambda t=delay: time.sleep(int(t) / 1000)
    return None


_COMMANDS = [
    _token_delay,
]


def run_command(token):
    for checker in _COMMANDS:
        cmd = checker(token)
        if callable(cmd):
            cmd()
            return True
    return False


def _tool(cmd, *args):
    cmd = [cmd]
    cmd.extend(args)
    subprocess.run(cmd)


def xdotool_type(delay, arg):
    _tool("xdotool", "type", "--delay", delay, "--", arg)


def xdotool_key(delay, key):
    _tool("xdotool", "key", "--clearmodifiers", "--delay", delay, "--", key)


def ydotool_type(delay, arg):
    _tool("ydotool", "type", "--key-delay", delay, "--", arg)


def ydotool_key(delay, key):
    _tool("ydotool", "key", "--key-delay", delay, "--", key)


def wtype_type(delay, arg):
    def _cmd(ch):
        if ch == "-":
            return ["-k", "minus"]
        return [ch]

    args = []
    for ch in arg:
        args.extend(_cmd(ch))
        args.extend(["-s", delay])

    _tool("wtype", *args)


def wtype_key(delay, key):
    _tool("wtype", "-k", key, "-s", delay)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2019 Michał Góral.

import sys
import os
import argparse
import subprocess
import time
import fnmatch
import shlex
import string
import secrets
import collections


class CommandError(Exception):
    pass


class ArgumentParserError(Exception):
    pass


class ThrowingArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        raise ArgumentParserError(message)


def prepare_command_parser():
    cp = ThrowingArgumentParser(prog="", add_help=False)
    sp = cp.add_subparsers(required=True)

    parsers = {}
    parsers[None] = cp

    # helper function which automatically creates help-friendly parsers
    def add_parser(command, *a, **kw):
        kw["add_help"] = False

        descr = kw.get("description")
        if descr:
            kw["description"] = "{}\n\n{}".format(kw["help"], descr)
        else:
            kw["description"] = kw["help"]

        parser = sp.add_parser(command, *a, **kw)
        parsers[command] = parser
        return parser

    ######### open
    open_sp = add_parser("open", help="Change currently opened database.")
    open_sp.add_argument("filepath", help="path to database file.")
    open_sp.set_defaults(func=open_)

    ######### unlock
    unlock_sp = add_parser("unlock", help="Unlock currently opened database.")
    unlock_sp.add_argument(
        "--keyfile", default="", help="key file used for unlocking database"
    )
    unlock_sp.set_defaults(func=unlock)

    ######### lock
    lock_sp = add_parser("lock", help="Lock a database.")
    lock_sp.set_defaults(func=lock)

    ######### db
    db_sp = add_parser("db", help="Query opened database info.")
    db_sp.set_defaults(func=db)

    ######### ls
    ls_sp = add_parser("ls", help="List contents of database.")
    ls_sp.add_argument(
        "glob",
        nargs="?",
        default="*",
        help="display only entries which match glob expression",
    )
    ls_sp.set_defaults(func=ls)

    ######### show
    show_sp = add_parser(
        "show", help="Show contents of entry.", description="Search is case-sensitive."
    )
    show_sp.add_argument("path", help="path which should be shown")
    show_sp.add_argument("fields", nargs="*", help="only display certain fields")
    show_sp.add_argument(
        "-n",
        "--no-field-name",
        action="store_true",
        help="hide field name when printing entry fields.",
    )
    show_sp.set_defaults(func=show)

    ######### add/edit
    add_sp = add_parser(
        "add",
        help="Add a new entry if it doesn't exist yet.",
        description="New entry can be in form 'group/name'. In that case, "
        "it will be added as a member of existing group or the new "
        "group will be created",
    )
    edit_sp = add_parser("edit", help="Edit existing entry")

    for spr in (add_sp, edit_sp):
        spr.add_argument("path", help="entry path")
        spr.add_argument("-u", "--username", help="username")
        spr.add_argument("-p", "--password", help="password")

        if spr is edit_sp:
            spr.add_argument(
                "--askpass",
                action="store_true",
                help="interactively ask for password; input will be hidden",
            )

        spr.add_argument("-U", "--url", help="URL")
        spr.add_argument(
            "-n",
            "--note",
            action="append",
            dest="notes",
            help="add a note. One note takes one line. Many notes can be added "
            "by using -n more than once",
        )

        if spr is edit_sp:
            spr.add_argument(
                "-N",
                "--delete-note",
                dest="delnotes",
                metavar="NOTES",
                help="delete note. Notes are accessed by their index. Range of "
                "notes can be deleted by passing e.g. '-N 2-5', or all notes can "
                "be deleted by passing '-N *'",
            )

        spr.add_argument(
            "-s",
            "--autotype-sequence",
            metavar="KEYS",
            help="key sequence used for autotyping",
        )

        sp_gen = spr.add_argument_group("password generation")
        sp_gen.add_argument(
            "-g",
            "--pw-gen",
            action="store_true",
            help="enable password generation if --password is not set. By "
            "default password is generated from ASCII letters and digits",
        )
        sp_gen.add_argument(
            "--letters",
            dest="charset",
            action="append_const",
            const=string.ascii_letters,
            help="use ASCII letters in generated password: a-z, A-Z",
        )
        sp_gen.add_argument(
            "--digits",
            dest="charset",
            action="append_const",
            const=string.digits,
            help="use digits in generated password: 0-9",
        )
        sp_gen.add_argument(
            "--punctuation",
            dest="charset",
            action="append_const",
            const=string.punctuation,
            help=f"use punctuation symbols in generated password: !#,.:; etc.",
        )
        sp_gen.add_argument(
            "--characters",
            help="characters which should be used for password generation. "
            "Setting this option overrides usage of --letters, --digits and "
            "--punct.",
        )
        sp_gen.add_argument(
            "-l",
            "--length",
            metavar="N",
            type=int,
            default=20,
            help="length of generated password (default: 20)",
        )

    add_sp.set_defaults(func=add)
    edit_sp.set_defaults(func=edit)

    ######### delete
    del_sp = add_parser("delete", help="Delete entry from database")
    del_sp.add_argument("paths", nargs="+", help="path of entry to remove")
    del_sp.set_defaults(func=delete)

    ######### autotype
    at_sp = add_parser(
        "autotype",
        help="Auto-type sequence of entry fields.",
        description="This simulates keypresses to any currently open window. "
        "It's particularily useful when kpsh is run from a script "
        "or keypress in non-interactive mode (`-c` switch). If "
        "`-s` is given, it will be used as auto-type sequence. "
        "Otherwise sequence defined for selected entry will be "
        "used or the default one if there is none (`-d`).",
    )
    at_sp.add_argument("path", help="path of entry to auto-type")
    at_sp.add_argument("-s", "--sequence", help="override auto-type sequence")
    at_sp.add_argument(
        "-d",
        "--default",
        default="{USERNAME}{TAB}{PASSWORD}{ENTER}",
        help="default auto-type sequence used when entry doesn't specify "
        "sequence itself.",
    )
    at_sp.add_argument(
        "-D", "--delay", type=int, default=40, help="delay beteen simulated keypresses"
    )
    at_sp.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="force auto-type for entries for which auto-type was disabled",
    )
    at_sp.add_argument(
        "-b",
        "--backend",
        choices=("xdotool", "ydotool", "wtype"),
        help="force usage of backend program for typing",
    )
    at_sp.add_argument(
        "-B",
        "--backend-cmd",
        help="command which will be run before autotype to detect backend. It "
        "should print the backend name (see --backend) to the standard "
        "output.",
    )
    at_sp.set_defaults(func=autotype)

    ######### exit
    exit_sp = add_parser("exit", help="Exit shell.")
    exit_sp.set_defaults(func=exit)

    ######### echo
    echo_sp = add_parser("echo", help="Display a message.")
    echo_sp.add_argument("message", nargs="*", help="message to be displayed")
    echo_sp.set_defaults(func=echo)

    ######### sleep
    sleep_sp = add_parser(
        "sleep",
        help="Sleep for a given number of seconds.",
        description="Seconds might be a floating number when fractions of "
        "second are needed.",
    )
    sleep_sp.add_argument("secs", type=float, help="seconds to sleep")
    sleep_sp.set_defaults(func=sleep)

    ######### help
    help_sp = add_parser("help", help="Show help for any message.")
    help_sp.add_argument("command", nargs="?")
    help_sp.set_defaults(func=lambda *a, parsers=parsers: help_(*a, parsers))

    return cp, parsers


def tokenize(seq):
    tokens = []

    i = 0
    while i < len(seq):
        start = seq.find("{", i)

        if start != -1:
            if i < start:
                tokens.append(seq[i:start])

            end = seq.find("}", start)
            nend = end + 1

            if end == -1:
                end = len(seq) - 1
            elif end == start + 1 and len(seq) > nend and seq[nend] == "}":  # {}}
                end = nend

            tokens.append(seq[start : end + 1])
            i = end + 1
        else:
            tokens.append(seq[i:])
            i = len(seq)

    return tokens


def autotype_environment(backend, backend_cmd, ioh):
    def _xdotool():
        from kpsh.autotype.xdotoolkeys import XDOTOOL_KEYS
        from kpsh.autotype.commands import xdotool_type, xdotool_key

        return xdotool_type, xdotool_key, XDOTOOL_KEYS

    def _ydotool():
        from kpsh.autotype.ydotoolkeys import YDOTOOL_KEYS
        from kpsh.autotype.commands import ydotool_type, ydotool_key

        return ydotool_type, ydotool_key, YDOTOOL_KEYS

    def _wtype():
        from kpsh.autotype.wtypekeys import WTYPE_KEYS
        from kpsh.autotype.commands import wtype_type, wtype_key

        return wtype_type, wtype_key, WTYPE_KEYS

    # backend_cmd will not return anything; instead it will set backend
    # variable. This way we'll avoid validating backend_cmd's output twice
    if not backend and backend_cmd:
        cmd = shlex.split(backend_cmd)
        try:
            cp = subprocess.run(cmd, capture_output=True, text=True)
            if cp.returncode == 0:
                backend = cp.stdout.strip().lower()
        except:
            ioh.eprint("Failed to run backend-cmd: {}".format(backend_cmd))
            pass

    if backend and backend == "xdotool":
        return _xdotool()
    elif backend and backend == "ydotool":
        return _ydotool()
    elif backend and backend == "wtype":
        return _wtype()
    elif os.environ.get("XDG_SESSION_TYPE", "") == "wayland":
        return _ydotool()
    return _xdotool()


def generate_password(characters, length):
    password = "".join(secrets.choice(characters) for i in range(length))
    return password


def askpass(ioh, prompt="Password"):
    return ioh.prompt("{}: ".format(prompt), is_password=True)


def autotype(kp, args, ioh):
    from kpsh.autotype.placeholders import replace_placeholder
    from kpsh.autotype.commands import run_command

    delay = str(args.delay)
    entry = _get(args.path, kp)

    if not entry.autotype_enabled and not args.force:
        ioh.eprint(
            "Autotype disabled for {}. " "Use -f to force autotype.".format(args.path)
        )
        return

    sequence = args.sequence if args.sequence else entry.autotype_sequence
    if not sequence:
        sequence = args.default

    typecmd, keycmd, TOOL_KEYS = autotype_environment(
        args.backend, args.backend_cmd, ioh
    )

    for token in tokenize(sequence):
        if token.startswith("{") and token.endswith("}"):
            if run_command(token):
                continue

            placeholder = replace_placeholder(entry, token)
            if placeholder is not None:
                typecmd(delay, placeholder)
                continue

            specialkey = TOOL_KEYS.get(token)
            if specialkey is not None:
                keycmd(delay, specialkey)
                continue

            ioh.eprint("Unsupported keyword: {}".format(token))
        else:
            typecmd(delay, token)


def notes_deleter(spec):
    def _delete_all(notes):
        return []

    # range is inclusive on both sides, so for example range 1-3 will delete
    # notes 1, 2 and 3
    def _delete_range(notes, start, end):
        return [note for i, note in enumerate(notes) if (i < start or i > end)]

    if spec == "*":
        return _delete_all

    if "-" in spec:
        start, _, end = spec.partition("-")
        start = int(start.strip())
        end = int(end.strip())
    else:
        start = int(spec.strip())
        end = start

    if start <= 0 or end <= 0:
        raise ValueError("Invalid range (must be > 0): {} - {}".format(start, end))

    return lambda notes: _delete_range(notes, start - 1, end - 1)


def show(kp, args, ioh):
    def _print_single(spec, attr):
        ioh.print(spec.format(name=attr.name, value=attr.value))

    def _print_multi(spec, attr):
        for i, elem in enumerate(attr.value.splitlines()):
            indexed_name = "{name}[{i}]".format(name=attr.name, i=i + 1)
            ioh.print(spec.format(name=indexed_name, value=elem))

    entry = _get(args.path, kp)
    Attr = collections.namedtuple("Data", ("name", "value", "printer"))

    attrs = [
        Attr("path", args.path, _print_single),
        Attr("username", entry.username, _print_single),
        Attr("password", entry.password, _print_single),
        Attr("url", entry.url, _print_single),
        Attr("autotype_sequence", entry.autotype_sequence, _print_single),
        Attr("notes", entry.notes, _print_multi),
    ]

    def find_attr(name):
        for attr in attrs:
            if attr.name == name:
                return attr
        return None

    fields = args.fields if args.fields else [a.name for a in attrs]

    for field in fields:
        attr = find_attr(field)
        if attr is None:
            ioh.eprint("Unknown attribute: {}".format(field))
            continue

        if attr.value is None:
            continue

        spec = "{value}" if args.no_field_name else "{name}: {value}"
        attr.printer(spec, attr)


def _set_entry_fields(entry, **fields):
    edited = False
    for attr, val in fields.items():
        if val is not None:
            setattr(entry, attr, val)
            edited = True
    return edited


def add(kp, args, ioh):
    if args.path in kp.entries:
        raise CommandError("entry already exists: {}".format(args.path))

    groupname, _, title = args.path.rpartition("/")
    group = kp.add_group(groupname) if groupname else kp.root_group

    if not group:
        raise CommandError("Failed adding a new group: {}".format(groupname))

    entry_args = {"username": args.username if args.username else ""}
    if args.password:
        entry_args["password"] = args.password
    elif args.pw_gen:
        characters = None
        if args.characters:
            characters = args.characters
        elif args.charset:
            characters = "".join(args.charset)
        else:
            characters = string.ascii_letters + string.digits
        entry_args["password"] = generate_password(characters, args.length)
    else:
        prompt = "Password for {}".format(title)
        entry_args["password"] = askpass(ioh, prompt)

    if not entry_args["password"]:
        raise CommandError("Setting a password is necessary for a new entry")

    try:
        with kp.no_reload():
            entry = kp.add_entry(group, title, **entry_args)
    except Exception as e:
        raise CommandError(str(e))

    notes = "\n".join(args.notes) if args.notes else None
    _set_entry_fields(entry, url=args.url, notes=notes)

    # This overwrites autotype_sequence set by PyKeePass, which incorrectly
    # sets it to a string 'None'.
    # See: https://github.com/libkeepass/pykeepass/issues/284
    entry.autotype_sequence = args.autotype_sequence

    kp.save()


def edit(kp, args, ioh):
    entry = _get(args.path, kp)

    password = None
    if args.password:
        password = args.password
    elif args.askpass:
        prompt = "New password for {}".format(entry.title)
        password = askpass(ioh, prompt)
    elif args.pw_gen:
        characters = None
        if args.characters:
            characters = args.characters
        elif args.charset:
            characters = "".join(args.charset)
        else:
            characters = string.ascii_letters + string.digits
        password = generate_password(characters, args.length)

    old_notes = entry.notes or ""
    old_notes_list = old_notes.splitlines()

    if args.delnotes:
        try:
            deleter = notes_deleter(args.delnotes)
            old_notes_list = deleter(old_notes_list)
        except (ValueError, TypeError):
            raise CommandError("Invalid value for -N: {}".format(args.delnotes))

    new_notes = args.notes or []
    notes = old_notes_list + new_notes
    notes = "\n".join(notes) if notes else ""

    edited = _set_entry_fields(
        entry,
        username=args.username,
        password=password,
        url=args.url,
        notes=notes,
        autotype_sequence=args.autotype_sequence,
    )

    if edited:
        kp.save()


def delete(kp, args, ioh):
    deleted = False
    with kp.no_reload():
        for path in args.paths:
            entry = kp.delete_entry(path)
            deleted |= bool(entry)

    if deleted:
        kp.save()


def echo(kp, args, ioh):
    ioh.print(*args.message)


def sleep(kp, args, ioh):
    time.sleep(args.secs)


def ls(kp, args, ioh):
    for path in fnmatch.filter(kp.iter_paths(), args.glob):
        ioh.print(path)


def help_(kp, args, ioh, parsers):
    parser = parsers.get(args.command)
    if parser is None:
        ioh.eprint("No such command: {}".format(args.command))
        parser = parsers.get(None)

    parser.print_help()


def exit(kp, args, ioh):
    ioh.stop()


def open_(kp, args, ioh):
    fp = os.path.expanduser(args.filepath)
    kp.change_db(fp)


def unlock(kp, args, ioh):
    if not kp.locked:
        return

    kf = os.path.expanduser(args.keyfile) if args.keyfile else None
    kp.change_credentials(keyfile=kf)
    kp.unlock()


def lock(kp, args, ioh):
    kp.lock()


def db(kp, args, ioh):
    ioh.print(kp.db)
    ioh.print("Locked: {}".format(kp.locked))


def _get(path, kp):
    entry = kp.entries.get(path)
    if not entry:
        raise CommandError("Entry not found: {}".format(path))
    return entry



# SPDX-License-Identifier: GPL-3.0-or-later

# Copyright (C) 2019 Michał Góral.

import os
import sys
import argparse
import tempfile
import signal

from kpsh.io import Interactive, NonInteractive, Daemon, Arguments

from kpsh import __version__ as _version
from kpsh.db import DelayedPyKeePass
from kpsh.commands import prepare_command_parser


class Cleanup:
    def __init__(self):
        signal.signal(signal.SIGINT, self.sysexit)
        signal.signal(signal.SIGTERM, self.sysexit)

    def sysexit(self, signum, frame):
        raise SystemExit(signum)


def default_socket_path():
    tmp = tempfile.gettempdir()
    fname = "kpsh-{}.sock".format(os.getuid())
    return os.path.join(tmp, fname)


def prepare_args():
    ap = argparse.ArgumentParser(description="KeePass database shell access.")
    ap.add_argument("db", nargs="?", help="path to KeePass database.")

    ap.add_argument("--password", default=None, help="Database password.")
    ap.add_argument(
        "--pw-cmd",
        default=None,
        help="Password will be obtained from the output of this command.",
    )
    ap.add_argument("--keyfile", default=None, help="Key file for unlocking database.")
    ap.add_argument("--pinentry", help="Command used to run pinentry.")
    ap.add_argument(
        "-c",
        "--command",
        action="append",
        help="Command to execute. If command arguments contain spaces, they "
        "must be enclosed in double quotes. With this switch, kpsh will be "
        "started in non-interactive mode. A list of available commands can "
        "be found by running 'kpsh -c help'.",
    )
    ap.add_argument(
        "--prompt",
        default='<style fg="ansiblue">{}</style>> ',
        help="Text used by shell for prompt.",
    )
    ap.add_argument(
        "-d",
        "--daemon",
        action="store_true",
        help="Start as a daemon listening on a socket given by --socket-path",
    )
    ap.add_argument(
        "-s",
        "--socket-path",
        default=default_socket_path(),
        help="Path to the socket which will be created in daemon mode "
        "(default: /tmp/kpsh-$UID.sock).",
    )
    ap.add_argument(
        "--version", action="version", version="%(prog)s {}".format(_version)
    )
    return ap.parse_args()


def main():
    cl = Cleanup()
    args = prepare_args()

    cmd_parser, parsers = prepare_command_parser()

    if args.command:
        ioh = Arguments(args.command)
    elif args.daemon:
        ioh = Daemon(args.socket_path)
    elif sys.stdin.isatty():
        ioh = Interactive(args.prompt, parsers)
    else:
        ioh = NonInteractive()

    kp = DelayedPyKeePass(args, ioh)
    return ioh.run(kp, cmd_parser)



import os
import string

import pytest

from kpsh.commands import CommandError


@pytest.fixture
def kpbig(kp_factory):
    return kp_factory("db_argon2_kdbx4_pass_big.kdbx", password="foobar")


def test_add_dont_override_existing_entry(kpbig, th):
    with pytest.raises(CommandError):
        th.cmd(f"add foo", kpbig.kp, kpbig.ioh)


@pytest.mark.parametrize("path", ["newentry", "group/newentry", "newgroup/entry"])
def test_add_username_password(kpbig, th, path):
    user = "myuser"
    pwd = "mypass"
    th.cmd(f"add {path} --username {user} --password {pwd}", kpbig.kp, kpbig.ioh)
    th.assert_fields(kpbig.kp, path, username=user, password=pwd)


def test_add_password_interactively(kpbig, th):
    kpbig.ioh.prompt.return_value = "newentry_password"
    th.cmd(f"add newentry", kpbig.kp, kpbig.ioh)
    th.assert_fields(kpbig.kp, "newentry", username="", password="newentry_password")


def test_add_generate_password(kpbig, th):
    th.cmd(f"add newentry --pw-gen -l 32", kpbig.kp, kpbig.ioh)
    fields = th.fields("newentry", kpbig.kp)
    assert fields
    assert len(fields["password"]) == 32


@pytest.mark.parametrize(
    "arg,chset",
    [
        ("--letters", string.ascii_letters),
        ("--digits", string.digits),
        ("--punctuation", string.punctuation),
        ("--characters abc", ["a", "b", "c"]),
    ],
)
def test_add_generate_password_character_sets(kpbig, th, arg, chset):
    th.cmd(f"add newentry --pw-gen -l 100 {arg}", kpbig.kp, kpbig.ioh)

    fields = th.fields("newentry", kpbig.kp)
    assert fields
    assert len(fields["password"]) == 100

    for ch in fields["password"]:
        assert (
            ch in chset
        ), f"'{ch}' is in set of letters allowed for generated password"


def test_add_notes(kpbig, th):
    pwd = "pwd"
    th.cmd(
        f"add newentry -p {pwd} --note 'note 1' --note 'note 2'", kpbig.kp, kpbig.ioh
    )
    th.assert_fields(
        kpbig.kp, "newentry", username="", password=pwd, notes=["note 1", "note 2"]
    )


def test_add_simple_fields(kpbig, th):
    pwd = "pwd"
    th.cmd(
        f"add newentry -p {pwd} --url https://example.com --autotype-sequence {{TAB}}",
        kpbig.kp,
        kpbig.ioh,
    )
    th.assert_fields(
        kpbig.kp,
        "newentry",
        username="",
        password=pwd,
        url="https://example.com",
        autotype_sequence="{TAB}",
    )


def test_add_detect_file_modification(kpbig, th, kp_open):
    def modify_db_and_answer_password(text, is_password):
        assert is_password is True
        db = kpbig.args.db
        curr = os.stat(db)
        os.utime(db, (curr.st_atime, curr.st_mtime + 100))
        return "pwd"

    kpbig.ioh.prompt.side_effect = modify_db_and_answer_password
    th.cmd(f"add newentry", kpbig.kp, kpbig.ioh)

    assert kpbig.kp.db != kpbig.args.db
    assert os.path.exists(kpbig.args.db)
    assert os.path.exists(kpbig.kp.db)

    oldkp = kp_open(kpbig.args.db, password="foobar")
    assert th.fields("newentry", oldkp.kp) is None

    newkp_fields = th.fields("newentry", kpbig.kp)
    assert newkp_fields
    assert newkp_fields["password"] == "pwd"



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2019 Michał Góral.

import os
import sys
import subprocess
import operator
import collections
import shutil
import contextlib
from datetime import datetime as dt


class DatabaseError(Exception):
    pass


# Delays importing of PyKeePass which by itself adds ~150 ms to startup...
class DelayedPyKeePass:
    def __init__(self, args, ioh):
        self._db = None
        self._ioh = ioh
        if args.db:
            self.change_db(args.db)

        self._password = args.password
        self._keyfile = args.keyfile

        self._pw_cmd = args.pw_cmd
        self._pinentry = args.pinentry

        self._kp = None
        self._mtime = None
        self._entries = None
        self._reload_guard = 0

    def _getpass(self):
        if self._password is not None:
            yield self._password
        if self._pw_cmd is not None:
            yield readpass(self._pw_cmd)
        if (
            self._password is None
            and self._pw_cmd is None
            and self._keyfile is not None
        ):
            # password=None supports Composite Master Key with only a keyfile
            # https://keepass.info/help/base/keys.html
            yield None
        yield askpass(self._db, self._pinentry, self._ioh)

    def unlock(self):
        if self._kp and self._reload_guard > 0:
            return True

        if self._kp and self._mtime == self._read_mtime():
            return True
        elif self._kp:
            # If e.g. database password was changed, this might fail. That's OK,
            # but we must allow user to provide a password in a second,
            # "ordinary" database read.
            if self._read_db([self._kp.password], self._kp.keyfile):
                return True
            else:
                self._ioh.eprint(
                    "Database was modified externally and must be reloaded, "
                    "but auto-reload failed. Re-obtaining password."
                )

        if not self._db:
            self._ioh.eprint("No database is selected.")
            return False

        if not self._read_db(self._getpass, self._keyfile):
            self._ioh.eprint("Failed unlocking database {}.".format(self._db))
            return False
        return True

    def lock(self):
        self._kp = None
        self._entries = None
        self._password = None
        self._keyfile = None

    def change_db(self, db):
        self.lock()
        if not os.path.isfile(db) and os.path.exists(db):
            self._ioh.eprint("Invalid database: '{}'".format(db))
            return
        if not os.path.exists(db):
            self._ioh.eprint(
                "Database doesn't exist and will be created on the first use: "
                "{}.".format(db)
            )
        self._db = db

    def change_credentials(self, password=None, keyfile=None):
        if not self.locked:
            raise DatabaseError(
                "Database must be locked before changing credentials "
                "used for unlocking."
            )

        self._password = password
        self._keyfile = keyfile

    @property
    def locked(self):
        return self._kp is None

    @property
    def db(self):
        return self._db

    @property
    def kp(self):
        self.unlock()
        return self._kp

    @property
    def entries(self):
        if not self.kp:
            return collections.OrderedDict()

        if self._entries is None:
            self._entries = collections.OrderedDict(
                ("/".join(e.path), e)
                for e in sorted(self.kp.entries, key=operator.attrgetter("path"))
            )

        return self._entries

    @property
    def paths(self):
        return [e for e in self.entries.keys()]

    def iter_paths(self):
        for entry in self.entries:
            yield entry

    def add_group(self, path):
        kp = self.kp

        if not kp:
            return None

        if not path:
            raise ValueError("Unexpected empty path")

        path = path.split("/")
        curr = kp.root_group

        for part in path:
            found = kp.find_groups(name=part, group=curr, recursive=False, first=True)
            if found:
                curr = found
            else:
                curr = kp.add_group(curr, part)

        return curr

    @property
    def root_group(self):
        if not self.kp:
            return None
        return self.kp.root_group

    def add_entry(self, group, *a, **kw):
        entry = self.kp.add_entry(group, *a, **kw)
        self.reload_entries()
        return entry

    def delete_entry(self, path):
        entry = self.entries.get(path)
        if entry:
            self.kp.delete_entry(entry)
            self.reload_entries()
            return entry
        return None

    def save(self):
        assert self._kp, "database not opened before save() call"
        assert (
            self._reload_guard == 0
        ), "save() can't be called when no_reload is active"

        if self._mtime != self._read_mtime():
            newf = next_file(self._kp.filename)
            self._kp.save(newf)
            self._kp.filename = newf
            self._db = newf
            self._mtime = self._read_mtime()
            self._ioh.eprint(
                "Database was modified externally. To prevent data loss, "
                "changes were saved to the new database: {}".format(newf)
            )
            return

        self._kp.save()
        self._mtime = self._read_mtime()

    @contextlib.contextmanager
    def no_reload(self):
        try:
            self._reload_guard += 1
            yield
        finally:
            self._reload_guard -= 1

    def reload_entries(self):
        self._entries = None

    def _read_mtime(self):
        return os.stat(self._db).st_mtime

    def _read_db(self, password_iter, keyfile):
        from pykeepass import PyKeePass, create_database

        if callable(password_iter):
            password_iter = password_iter()

        def _open_db(password, keyfile):
            if os.path.exists(self._db):
                kp = PyKeePass(self._db, password=password, keyfile=keyfile)
            else:
                kp = create_database(self._db, password=password, keyfile=keyfile)
                pwfmt = "<redacted>" if password else "<empty>"
                kwfmt = keyfile if keyfile else "<none>"
                self._ioh.eprint(
                    "Database created: {db} (password: {pw}, keyfile: {kf})".format(
                        db=self._db, pw=pwfmt, kf=kwfmt
                    )
                )
            return kp

        kp = None
        for pw in password_iter:
            try:
                kp = _open_db(password=pw, keyfile=keyfile)
            except Exception as e:
                pass
            else:
                self._mtime = self._read_mtime()
                break

        self._kp = kp
        self._entries = None
        return kp is not None


def askpass(db, pinentry, ioh):
    if pinentry:
        return _askpass_pinentry(db, pinentry)
    return _askpass_prompt(ioh)


def readpass(cmd):
    cp = subprocess.run(cmd, capture_output=True, text=True, shell=True)
    if cp.returncode != 0 or not cp.stdout:
        return None
    return cp.stdout


def next_file(path):
    while True:
        ts = dt.now().timestamp()
        nextf = "{}~{}".format(path, ts)
        if not os.path.exists(nextf):
            return nextf


def _askpass_pinentry(db, pinentry):
    pein = "setdesc Enter password for {}\ngetpin\n".format(db)
    cp = subprocess.run(pinentry, input=pein, capture_output=True, text=True)

    if cp.returncode != 0 or not cp.stdout:
        return None

    lines = cp.stdout.splitlines()
    dlines = [line for line in lines if line.startswith("D ")]
    if not dlines:
        return None

    passline = dlines[0]
    return passline.partition(" ")[-1]


def _askpass_prompt(ioh):
    return ioh.prompt("Database password: ", is_password=True)



import os
import pytest

from kpsh.db import DelayedPyKeePass


@pytest.fixture
def kp(kp_factory):
    def opener(**args):
        return kp_factory("db_argon2_kdbx4_pass.kdbx", **args)

    return opener


def test_states(kp):
    kp = kp(password="foobar").kp
    assert kp.locked
    assert kp.unlock()
    assert not kp.locked


def test_fail_unlock(kp):
    kp = kp().kp
    assert not kp.unlock()


def test_unlock_of_unlocked_db(kp):
    kpdata = kp()
    kp = kpdata.kp
    ioh = kpdata.ioh

    ioh.prompt.return_value = "foobar"
    assert kp.unlock()

    ioh.prompt.return_value = None
    assert kp.unlock()


def test_auto_reload(kp):
    kpdata = kp()
    kp = kpdata.kp
    ioh = kpdata.ioh
    args = kpdata.args

    ioh.prompt.return_value = "foobar"

    assert kp.unlock()
    assert kp.kp

    # Let's change the passwords so we can verify that auto-reloading fails
    kp.kp.password = None
    ioh.prompt.return_value = None

    # Modify mtime of our database
    curr = os.stat(args.db)
    os.utime(args.db, (curr.st_atime, curr.st_mtime + 100))

    assert not kp.kp  # failed auto-unlocking because we removed password

    ioh.prompt.return_value = "foobar"
    assert kp.kp



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2019 Michał Góral.

_PLACEHOLDERS = {
    "{TITLE}": lambda e: e.title,
    "{USERNAME}": lambda e: e.username,
    "{URL}": lambda e: e.url,
    "{PASSWORD}": lambda e: e.password,
    "{NOTES}": lambda e: e.notes,
    "{PLUS}": lambda _: "+",
    "{PERCENT}": lambda _: "%",
    "{CARET}": lambda _: "^",
    "{TILDE}": lambda _: "~",
    "{LEFTPAREN}": lambda _: "(",
    "{RIGHTPAREN}": lambda _: ")",
    "{LEFTBRACE}": lambda _: "{",
    "{RIGHTBRACE}": lambda _: "}",
    "{AT}": lambda _: "@",
    "{+}": lambda _: "+",
    "{%}": lambda _: "%",
    "{^}": lambda _: "^",
    "{~}": lambda _: "~",
    "{(}": lambda _: "(",
    "{)}": lambda _: ")",
    "{[}": lambda _: "[",
    "{]}": lambda _: "]",
    "{{}": lambda _: "{",
    "{}}": lambda _: "}",
}


def replace_placeholder(entry, token):
    fn = _PLACEHOLDERS.get(token)
    if callable(fn):
        return fn(entry)
    return None



import os
import glob
import re

from prompt_toolkit.completion import Completer, Completion

_OPT_PATTERN = re.compile(r"([a-zA-Z0-9_]+|[^a-zA-Z0-9_\s]+)")


class CommandCompleter(Completer):
    def __init__(self, kp, commands):
        self._commands = sorted(commands)
        self._kp = kp

    def get_completions(self, document, complete_event):
        text_before = document.text_before_cursor
        if " " in text_before:
            command = text_before.partition(" ")[0]
            fn = getattr(self, "_{}".format(command), None)
            if callable(fn):
                yield from fn(document)
        else:
            yield from self._complete_commands(document)

    def _complete_commands(self, document):
        text_before = document.text_before_cursor
        for command in self._commands:
            if command.startswith(text_before):
                yield Completion(command, -len(text_before))

    def _show(self, document):
        opts = ["--no-field-name"]
        yield from self._path_and_opts(document, opts)

    def _autotype(self, document):
        opts = [
            "--default::change default sequence",
            "--sequence::override sequence",
            "--force::force auto-type when disabled",
            "--delay::delay between keypresses",
        ]
        yield from self._path_and_opts(document, opts)

    def _add(self, document):
        opts = [
            "--username::username",
            "--password::password",
            "--url::URL",
            "--note::one line in notes",
            "--autotype-sequence::key sequence",
            "--pw-gen::generate password",
            "--letters::generate with letters",
            "--digits::generate with lude digits",
            "--punctuation::generate with punctuation",
            "--characters::list of characters for password",
            "--length::password length",
        ]
        yield from self._path_and_opts(document, opts)

    def _edit(self, document):
        opts = [
            "--username::username",
            "--password::password",
            "--askpass::ask for password",
            "--url::URL",
            "--note::one line in notes",
            "--autotype-sequence::key sequence",
            "--pw-gen::generate password",
            "--letters::generate with letters",
            "--digits::generate with lude digits",
            "--punctuation::generate with punctuation",
            "--characters::list of characters for password",
            "--length::password length",
        ]
        yield from self._path_and_opts(document, opts)

    def _delete(self, document):
        opts = []
        yield from self._path_and_opts(document, opts)

    def _help(self, document):
        word = document.get_word_before_cursor()
        for command in self._commands:
            if command.startswith(word):
                yield Completion(command, -len(word))

    def _open(self, document):
        yield from self._filepath(document)

    def _path_and_opts(self, document, opts):
        word = document.get_word_before_cursor(WORD=True)
        if word.startswith("-"):
            completions = sorted(opts)
        else:
            completions = self._kp.paths if not self._kp.locked else []

        for compl in completions:
            compl, _, meta = compl.partition("::")
            if not meta:
                meta = None
            if compl.startswith(word):
                yield Completion(compl, -len(word), display_meta=meta)

    def _filepath(self, document):
        word = document.get_word_before_cursor(WORD=True)

        home = os.path.expanduser("~")
        tilde_repl = word.startswith("~/")
        for compl in glob.iglob("{}*".format(os.path.expanduser(word))):
            meta = None
            if os.path.isfile(compl):
                meta = "file"
            elif os.path.isdir(compl):
                meta = "dir"

            if tilde_repl:
                compl = compl.replace(home, "~")

            yield Completion(compl, -len(word), display_meta=meta)



import pytest


@pytest.fixture
def kpbig(kp_factory):
    return kp_factory("db_argon2_kdbx4_pass_big.kdbx", password="foobar")


def test_delete(kpbig, th):
    expected = kpbig.kp.entries.copy()
    del expected["foo"]

    assert "foo" in kpbig.kp.entries
    th.cmd("delete foo", kpbig.kp, kpbig.ioh)
    assert kpbig.kp.entries == expected


def test_dont_delete_missing_entry(kpbig, th):
    expected = kpbig.kp.entries.copy()

    th.cmd(f"delete somethingwhichdoesntexist", kpbig.kp, kpbig.ioh)
    assert kpbig.kp.entries == expected


def test_delete_subgroup(kpbig, th):
    expected = kpbig.kp.entries.copy()
    del expected["other/foo"]

    th.cmd("delete other/foo", kpbig.kp, kpbig.ioh)
    assert kpbig.kp.entries == expected


def test_delete_many(kpbig, th):
    expected = kpbig.kp.entries.copy()
    del expected["foo"]
    del expected["other/foo"]
    del expected["group/subgroup/other entry"]

    th.cmd("delete foo other/foo 'group/subgroup/other entry'", kpbig.kp, kpbig.ioh)
    assert kpbig.kp.entries == expected



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['kpsh', 'kpsh.autotype']

package_data = \
{'': ['*']}

install_requires = \
['prompt_toolkit>=3.0.21,<4.0.0', 'pykeepass>=4.0.0,<5.0.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=1.0,<2.0']}

entry_points = \
{'console_scripts': ['kpsh = kpsh.app:main']}

scripts = \
['contrib/kpsh-client', 'contrib/kpsh-detect-backend-sway', 'contrib/kpsh-menu']

setup_kwargs = {
    'name': 'kpsh',
    'version': '1.0.0',
    'description': 'KeePass shell interface and daemon',
    'long_description': '# kpsh\n\nkpsh, or KeePass Shell, is a password manager and an interactive shell for\nworking directly with KeePass password database files.\n\n## Features\n\n- create, open, lock and unlock databases\n- add, edit and delete database entries\n- list contents of database\n- show contents of database entries and filter them by fields\n- autotype usernames and passwords or any sequences of entry fields (by\n  xdotool on X11 and ydotool on Wayland)\n- access all commands non-interactively via `-c` switch or by piping commands\n  directly to kpsh\n- tab-completion in interactive mode\n- daemon mode: open and unlock your database once and then quickly access\n  its contents from kpsh-client.\n- several built-in ways to obtain a password, which can be passed by argument,\n  typed directly to kpsh or through pinentry program program or fetched from a\n  provided command output\n- ships with highly customizable kpsh-menu script which performs any kpsh\n  command on entries selected by dmenu/rofi/fzf (e.g. autotype passwords\n  selected in dmenu/rofi)\n\n# Usage examples\n\nTypical session:\n\n```\n$ kpsh passwords.kdbx\n\npasswords.kdbx> ls\nPassword: ********\npersonal/bank\npersonal/login\npersonal/website\nwork/login\n\npasswords.kdbx> show work/login\npath: work/login\nusername: John Doe\npassword: jsdf7y8h8349yhj3h42\nnotes[1]: this is my work password\nnotes[2]: it\'s the best\n```\n\nGet a password from gpg-encrypted file (trailing newline, which isn\'t a part\nof password is trimmed):\n\n```\n$ gpg --encrypt -o masterpass.gpg -r mymail@example.com\n<type type type>\n^D\n$ kpsh passwords.kdbx --pw-cmd "gpg --decrypt masterpass.gpg | tr -d \'\\n\'"\n```\n\n... or from a keyring:\n\n```\n$ secret-tool store --label=\'keepass\' database passwords.kdbx\n$ kpsh passwords.kdbx --pw-cmd "secret-tool lookup database passwords.kdbx"\n```\n\nAutotype a user/password sequence:\n\n```\n$ kpsh passwords.kdbx --pw-cmd "secret-tool lookup database passwords.kdbx"\n                      -c autotype entry1\n```\n\n... or just a password, but a little faster:\n\n```\n$ kpsh passwords.kdbx --pw-cmd "secret-tool lookup database passwords.kdbx"\n                      -c "autotype -s {PASSWORD} -D 12 entry1"\n```\n\nRun as daemon (`-d`):\n\n```\n$ kpsh passwords.kdbx -d --pw-cmd "secret-tool lookup database passwords.kdbx" &\n$ kpsh-client ls\nentry1\nentry2\n$ kpsh-client autotype entry1\n```\n\nUse pinentry to get a password when database will be unlocked:\n\n```\n$ kpsh passwords.kdbx --pinentry /usr/bin/pinentry\n```\n\n## Installation\n\nUse [pipx][pipx]:\n\n```\n$ pipx install kpsh\n```\n\nOr directly pip:\n\n```\n$ pip install --user kpsh\n```\n\n  [pipx]: https://github.com/pypa/pipx\n',
    'author': 'Michal Goral',
    'author_email': 'dev@goral.net.pl',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://git.goral.net.pl/mgoral/keepass-shell',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'scripts': scripts,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import os
import string

import pytest

from kpsh.commands import CommandError


@pytest.fixture
def kpbig(kp_factory):
    return kp_factory("db_argon2_kdbx4_pass_big.kdbx", password="foobar")


@pytest.fixture
def shared_kpbig(shared_kp_factory):
    data = shared_kp_factory("db_argon2_kdbx4_pass_big.kdbx", password="foobar")
    data.ioh.print.reset_mock()
    return data


def test_edit_nothing(kpbig, th):
    expected = th.fields("foo", kpbig.kp).copy()
    th.cmd("edit foo", kpbig.kp, kpbig.ioh)
    assert th.fields("foo", kpbig.kp) == expected


@pytest.mark.parametrize(
    "field,switch,value",
    [
        ("username", "--username", "editeduser"),
        ("password", "--password", "editedpassword"),
        ("url", "--url", "editeturl"),
        ("autotype_sequence", "--autotype-sequence", "{TAB}{TAB}{TAB}"),
    ],
)
def test_edit_fields(kpbig, th, field, switch, value):
    expected = th.fields("foo", kpbig.kp).copy()
    assert expected[field] != value
    expected[field] = value

    th.cmd(f"edit foo {switch} {value}", kpbig.kp, kpbig.ioh)
    assert th.fields("foo", kpbig.kp) == expected


@pytest.mark.parametrize(
    "path", ["foo", "group/entry", "'group/subgroup/other entry'", "other/foo"]
)
def test_edit_fields_paths(kpbig, th, path):
    expected = th.fields(path, kpbig.kp).copy()
    expected["url"] = "newurl"

    th.cmd(f"edit {path} -U newurl", kpbig.kp, kpbig.ioh)
    assert th.fields(path, kpbig.kp) == expected


def test_edit_many_fields(kpbig, th):
    expected = th.fields("foo", kpbig.kp).copy()
    expected["username"] = "euser"
    expected["url"] = "eurl"
    expected["password"] = "epwd"

    th.cmd(f"edit foo -p epwd -U eurl -u euser ", kpbig.kp, kpbig.ioh)
    assert th.fields("foo", kpbig.kp) == expected


def test_edit_password_interactively(kpbig, th):
    expected = th.fields("foo", kpbig.kp).copy()
    expected["password"] = "editedpassword"
    kpbig.ioh.prompt.return_value = expected["password"]

    th.cmd(f"edit foo --askpass", kpbig.kp, kpbig.ioh)
    assert th.fields("foo", kpbig.kp) == expected


@pytest.mark.parametrize(
    "field,switch",
    [
        ("username", "--username"),
        ("url", "--url"),
        ("autotype_sequence", "--autotype-sequence"),
    ],
)
def test_edit_unset_fields(kpbig, th, field, switch):
    th.cmd(f"edit foo {switch} ''", kpbig.kp, kpbig.ioh)
    assert not th.fields("foo", kpbig.kp).get(field)


def test_edit_generate_password(kpbig, th):
    th.cmd(f"edit foo --pw-gen -l 32", kpbig.kp, kpbig.ioh)
    fields = th.fields("foo", kpbig.kp)
    assert len(fields["password"]) == 32


@pytest.mark.parametrize(
    "arg,chset",
    [
        ("--letters", string.ascii_letters),
        ("--digits", string.digits),
        ("--punctuation", string.punctuation),
        ("--characters abc", ["a", "b", "c"]),
    ],
)
def test_edit_generate_password_character_sets(kpbig, th, arg, chset):
    th.cmd(f"edit foo --pw-gen -l 100 {arg}", kpbig.kp, kpbig.ioh)

    fields = th.fields("foo", kpbig.kp)
    assert len(fields["password"]) == 100

    for ch in fields["password"]:
        assert (
            ch in chset
        ), f"'{ch}' is in set of letters allowed for generated password"


def test_edit_add_notes(kpbig, th):
    existing = th.fields("foo", kpbig.kp)["notes"]

    th.cmd("edit foo --note 'note 1'", kpbig.kp, kpbig.ioh)
    expected = existing + ["note 1"]
    assert th.fields("foo", kpbig.kp)["notes"] == expected

    th.cmd("edit foo --note 'note 2' --note 'note 3'", kpbig.kp, kpbig.ioh)
    expected = existing + ["note 1", "note 2", "note 3"]
    assert th.fields("foo", kpbig.kp)["notes"] == expected


@pytest.mark.parametrize(
    "pattern,expected",
    [
        ("1", ["2", "3", "4", "5"]),
        ("2", ["1", "3", "4", "5"]),
        ("5", ["1", "2", "3", "4"]),
        ("6", ["1", "2", "3", "4", "5"]),
        ("2-6", ["1"]),
    ],
)
def test_delete_notes(kpbig, th, pattern, expected):
    th.cmd("add tabula_rasa -p pwd -n1 -n2 -n3 -n4 -n5", kpbig.kp, kpbig.ioh)
    th.cmd(f"edit tabula_rasa --delete-note {pattern}", kpbig.kp, kpbig.ioh)
    assert th.fields("tabula_rasa", kpbig.kp)["notes"] == expected


@pytest.mark.parametrize("pattern", ["*", "1-5", "1-6"])
def test_delete_all_notes(kpbig, th, pattern):
    th.cmd("add tabula_rasa -p pwd -n1 -n2 -n3 -n4 -n5", kpbig.kp, kpbig.ioh)
    th.cmd(f"edit tabula_rasa --delete-note {pattern}", kpbig.kp, kpbig.ioh)
    assert "notes" not in th.fields("tabula_rasa", kpbig.kp)


@pytest.mark.parametrize(
    "pattern", ["0", "a", "0-1", "1-", "-", "-0", "-1", ",", "**", "1*"]
)
def test_delete_notes_incorrect_argument(shared_kpbig, th, pattern):
    th.cmd("edit foo -n1 -n2 -n3 -n4 -n5", shared_kpbig.kp, shared_kpbig.ioh)
    expected = th.fields("foo", shared_kpbig.kp).copy()

    cmd = f"edit foo --delete-note {pattern}"
    with pytest.raises(CommandError):
        th.cmd(cmd, shared_kpbig.kp, shared_kpbig.ioh)

    assert th.fields("foo", shared_kpbig.kp) == expected


def test_edit_add_delete_together(kpbig, th):
    existing = th.fields("foo", kpbig.kp)["notes"]
    assert existing

    th.cmd("edit foo -N * -n 'note 1'", kpbig.kp, kpbig.ioh)
    expected = ["note 1"]
    assert th.fields("foo", kpbig.kp)["notes"] == expected


def test_edit_detect_file_modification(kpbig, th, kp_open):
    def modify_db_and_answer_password(text, is_password):
        assert is_password is True
        db = kpbig.args.db
        curr = os.stat(db)
        os.utime(db, (curr.st_atime, curr.st_mtime + 100))
        return "pwd"

    expected_unchanged = th.fields("foo", kpbig.kp).copy()

    kpbig.ioh.prompt.side_effect = modify_db_and_answer_password
    th.cmd(f"edit foo --askpass", kpbig.kp, kpbig.ioh)

    assert kpbig.kp.db != kpbig.args.db
    assert os.path.exists(kpbig.args.db)
    assert os.path.exists(kpbig.kp.db)

    oldkp = kp_open(kpbig.args.db, password="foobar")
    assert th.fields("foo", oldkp.kp) == expected_unchanged

    newkp_fields = th.fields("foo", kpbig.kp)
    assert newkp_fields
    assert newkp_fields["password"] == "pwd"


@pytest.mark.parametrize("path", ["'missing path'", "missing", "foo/missing"])
def test_edit_not_existing_path(shared_kpbig, th, path):
    assert path not in shared_kpbig.kp.entries
    with pytest.raises(CommandError):
        th.cmd(f"edit {path} -p pwd", shared_kpbig.kp, shared_kpbig.ioh)
